import numpy as np

# Standard format
#    a0 a1 a2 ... ad
#    b0 b1 b2 ... bd <- 4 bits
# Format as in Quick ADC:
#    a0 b0 c0 ... p0 <- 16'th letter. Whole row is 64 bit
#    a1 b1 c1 ... p1
#    ...
#    ad bd cd ... pd
#    q0 r0 s0 ...    <- New chunk
# Loading data interleaved, since that's how shuffle reads it
#    a0 a1 b0 b1 ... h0 h1
#    i0 i1       ... p0 p1
#    ...
#    i(d-1) id   ... p(d-1) pd
#    q0 q1 r0 ...    <- New chunk
# Also remember that storage inside uint64 is right-to-left


# Table standard format
#    -a1- -a2- ... -a16-
#    -b1- -b2- ... -b16-
#    ...
#    {d}1 ...      {d}16
# Format for Quick ADC:
#    -a1- -a2- ... -a8-
#    -a9- -a10 ... -a16
#    -b1- -b2- ... -b8-
#    -b9- -b10 ... -b16
#    ...


def transform_data(data0):
    """
    Transform data from the standard format to the Quick-ADC format.

    The standard format is a 2D array with n rows and d columns.
    Each entry in the array is a 4-bit value, corresponding to a pointer to a cluster center.

    The Quick ADC format is a transformed version of the standard format, optimized
    for processing in the Quick-ADC algorithm. The data is first reshaped into groups of 16 rows,
    and interleaved by pairs of consecutive columns. Each group of 16 rows becomes a chunk.
    Within a chunk, data is stored in a 64-bit row-major order, with consecutive 4-bit values
    from 16 different rows combined into a single 64-bit value. The storage inside the 64-bit
    value is right-to-left, meaning that the first 4-bit value appears at the least significant
    bits of the 64-bit value.

    The transform_table function does the equivalent transformation for distance tables.
    For the tables, the standard format consists of 8-bit values in a 2D array. In the Quick ADC
    format, the tables are reshaped so that each row is divided into two equal parts and stacked
    one after the other.

    Parameters
    ----------
    data0 : numpy.ndarray
        A 2D input array with 4-bit entries. Its shape should be (n, d), where
        n is a multiple of 16 and d is even.

    Returns
    -------
    data : numpy.ndarray
        A transformed 2D array with dtype `np.uint64`. Its shape is (n // 16, d)
        since the 4-bit values have been packed in groups of 16.
    """
    n, d = data0.shape
    assert d % 2 == 0  # Because we load two rows at a time (128bits)
    assert np.all(data0 < 16) and np.all(0 <= data0)
    # Convert data to Quick-ADC format
    data = data0.reshape(n // 16, 16, d)
    data = data.transpose(0, 2, 1)
    # Interleave rows two by two
    data = (                                     # [0,1,2,3,4,5,6,7]
        data.reshape(n // 16, d // 2, 2, 16)     # [0,1,2,3], [4,5,6,7]
        .reshape(n // 16, d // 2, 32, order="F") # [0,4,1,5,2,6,3,7]
        .reshape(n // 16, d, 16)                 # [0,4,1,5], [2,6,3,7]
    )
    # Reversing last dimension for endianess
    data = data[:, :, ::-1]
    # Extra tests for required format
    assert data[0, 0, -1] == data0[0][0]
    assert data[0, 0, -2] == data0[0][1]
    assert data[0, 0, -3] == data0[1][0]
    assert data[0, 0, -4] == data0[1][1]
    assert data[0, 1, -1] == data0[8][0]
    assert data[0, 1, -2] == data0[8][1]
    if d > 2:
        assert data[0, 2, -1] == data0[0][2]
        assert data[0, 2, -2] == data0[0][3]
    assert data.shape == (n // 16, d, 16)
    # Converting last dimension to a single 64 bit number
    # data = np.frompyfunc(lambda x, y: (x << 4 | y), 2, 1).reduce(data, axis=2)
    shifts = np.arange(15, -1, -1, dtype=np.uint64) * 4
    data = (data << shifts).sum(axis=2, dtype=np.uint64)
    data = np.ascontiguousarray(data, dtype=np.uint64)
    assert data.shape == (n // 16, d)
    return data


def transform_tables(tables0):
    d, b = tables0.shape
    assert d % 2 == 0  # Because we load two rows at a time (128bits)
    assert b == 16
    assert tables0.dtype == np.uint8
    # Convert tables to 128 bit format
    tables = tables0.reshape(2 * d, 8)
    # Viewing takes care of swapping the byte order automatically
    tables = tables.view(np.uint64)[:, 0]
    assert tables.shape == (2 * d,)
    return tables


def unpack(transformed_data):
    """
    Unpack the transformed data back to its original format.

    Parameters
    ----------
    transformed_data : numpy.ndarray
        A 2D input array with dtype `np.uint64`. It is the output of the transform_data function.

    Returns
    -------
    data : numpy.ndarray
        A 2D array of the original format. Its shape should be (n, d), where
        d is the number of columns in the original data.
    """
    chunks, d = transformed_data.shape
    n = chunks * 16

    shifts = np.arange(15, -1, -1, dtype=np.uint64) * 4
    data = (transformed_data[..., np.newaxis] >> shifts) & 0xF

    data = data[:, :, ::-1]                               # [0,4,1,5], [2,6,3,7]
    data = data.reshape(chunks, d // 2, 32)               # [0,4,1,5,2,6,3,7]
    data = data.reshape(chunks, d // 2, 2, 16, order='F') # [0,1,2,3], [4,5,6,7]
    data = data.reshape(chunks, d, 16)                    # [0,1,2,3], [4,5,6,7]

    data = data.transpose(0, 2, 1)
    data = data.reshape(n, d)

    return data



import numpy as np
import sklearn.cluster
from collections import namedtuple

import warnings
from sklearn.exceptions import ConvergenceWarning

from ._transform import transform_data, transform_tables
from ._fast_pq import query_pq_sse, estimate_pq_sse, init_heap

warnings.simplefilter("error", category=ConvergenceWarning)

TransformedData = namedtuple("TransformedData", "size packed")

def pad1(arr, m):
    (s,) = arr.shape
    new_shape = (s + (-s) % m,)
    padded_arr = np.zeros(new_shape, dtype=arr.dtype)
    padded_arr[:s] = arr
    return padded_arr


def pad2(arr, m1, m2):
    s1, s2 = arr.shape
    new_shape = (s1 + (-s1) % m1, s2 + (-s2) % m2)
    padded_arr = np.zeros(new_shape, dtype=arr.dtype)
    padded_arr[:s1, :s2] = arr
    return padded_arr


def bottom_k(arr, k):
    if k >= len(arr):
        return np.arange(len(arr))
    return np.argpartition(arr, k)[:k]


def bottom_k_2d(arr, k):
    if k >= arr.shape[1]:
        return np.resize(np.arange(arr.shape[1]), arr.shape)
    return np.argpartition(arr, k, axis=1)[:, :k]


class FastPQ:
    def __init__(self, dims_per_block):
        """
        Initializes the FastPQ class with the specified number of dimensions per block.

        Parameters
        ----------
        dims_per_block : int
            The number of dimensions per block.
        """
        self.dims_per_block = dims_per_block
        self.centers = None  # Shape: (n_blocks, 16, dims_per_block)
        self.center_norms_sq = None  # Shape: (n_blocks, 16)

    def fit(self, data, verbose=False):
        """
        Fits the FastPQ model to the given data.

        Parameters
        ----------
        data : array-like
            The input data to fit the FastPQ model.
        verbose : bool, optional, default=False
            If True, prints additional information during the fitting process.

        Returns
        -------
        self : FastPQ
            The fitted FastPQ model.
        """
        _transformed_data = self.fit_transform(data, verbose)
        return self

    def fit_transform(self, data, verbose=False):
        assert data.size > 0, "Can't fit no data"
        true_n = data.shape[0]
        # SSE assumes the number of rows is divisible by 16.
        # It also needs the number of columns to be even, so we pad to a multiple
        # of 2 * self.dims_per_block.
        data = pad2(data, 16, 2 * self.dims_per_block)
        n, d = data.shape
        assert d % self.dims_per_block == 0
        assert (d // self.dims_per_block) % 2 == 0
        dpb = self.dims_per_block
        # We always use 16 clusters in FastPQ, since we want to use 4 bit SSE operations.
        cl = sklearn.cluster.KMeans(16, n_init=2)
        centers = []
        transformed = []
        for i in range(d // self.dims_per_block):
            if verbose:
                print(f"Fitting block {i}")
            try:
                cl.fit(data[:, i * dpb : (i + 1) * dpb])
            except ConvergenceWarning:
                pass
            # It doesn't give too much precision to do separate centers for each block,
            # but durnig queries we need seperate distance tables per block anyway, so
            # it doesn't cost us much.
            centers.append(cl.cluster_centers_.copy())
            # Might as well grab the labels (transformed column) while we have it.
            transformed.append(cl.labels_[:, None])
        self.centers = np.array(centers, dtype=np.float32)
        self.center_norms_sq = np.linalg.norm(centers, axis=2) ** 2

        labels = np.hstack(transformed).astype(np.uint8)
        return TransformedData(true_n, transform_data(labels))

    def transform(self, data):
        """
        Transforms the given data using the FastPQ model.

        Parameters
        ----------
        data : array-like
            The input data to transform using the FastPQ model.

        Returns
        -------
        tuple
            A tuple containing the transformed data and the number of true elements in the data.
        """
        assert self.centers is not None, "PQ has not been fitted"
        if data.size == 0:
            return data
        true_n = data.shape[0]
        data = pad2(data, 16, 2 * self.dims_per_block)
        n, d = data.shape
        assert n % 16 == 0
        parts = data.reshape(n, -1, self.dims_per_block)
        ips = np.einsum("ijk,jlk->ijl", parts, self.centers)  # Shape: (n, n_blocks, 16)
        labels = np.argmin(self.center_norms_sq - 2 * ips, axis=2)
        assert labels.shape == (n, d // self.dims_per_block)
        labels = np.array(labels, dtype=np.uint8)
        return TransformedData(true_n, transform_data(labels))

    def distance_table(self, q):
        """
        Computes the distance table for the given query vector q.

        Parameters
        ----------
        q : array-like
            The query vector.

        Returns
        -------
        _FastDistanceTable
            The FastDistanceTable object containing the computed distance table.
        """
        q = pad1(q, 2 * self.dims_per_block)
        dpb = self.dims_per_block
        n_blocks = q.size / dpb

        parts = q.reshape(-1, dpb)

        # Center the data in the range [-128, 128]
        dists = self.center_norms_sq - 2 * np.einsum("ijk,ik->ij", self.centers, parts)

        # We do this by shifting by the mean value and scaling by the nuber of blocks.
        # The idea is that we don't want to have an overflow as we add together
        # distances in uint8 format.
        # The median also works well here, maybe even a bit better, but it takes longer to compute.
        shift = np.median(dists)
        scale = 128 / (np.max(np.abs(dists - shift)) * np.sqrt(n_blocks))

        # Round to nearest integer towards zero.
        table = np.round((dists - shift) * scale)

        # The transformation doesn't care about the sign, so we just use uint
        table = table.astype(np.uint8)
        trans = transform_tables(table)
        return _FastDistanceTable(q, trans, shift, scale, signed=True)

    def udistance_table(self, q):
        """
        Computes the unsigned distance table for the given query vector q.
        Currently experimental.

        Parameters
        ----------
        q : array-like
            The query vector.

        Returns
        -------
        _FastDistanceTable
            The FastDistanceTable object containing the computed unsigned distance table.
        """
        q = pad1(q, 2 * self.dims_per_block)
        dpb = self.dims_per_block
        n_blocks = q.size / dpb
        parts = q.reshape(-1, dpb)
        dists = self.center_norms_sq - 2 * np.einsum("ijk,ik->ij", self.centers, parts)
        shift = np.min(dists)
        dists -= shift
        scale = 255 / (np.max(dists) * np.sqrt(n_blocks))
        table *= scale
        table = table.astype(np.uint8)
        trans = transform_tables(table)
        return _FastDistanceTable(q, trans, shift, scale, signed=False)


class _FastDistanceTable:
    def __init__(self, q, transformed_tables, mean, scale, signed):
        self.q = q
        self.tables = transformed_tables
        self.mean = mean
        self.scale = scale
        self.signed = signed

    def __repr__(self):
        return (
            f"FastDistanceTable(q={self.q}, tables={self.tables}, mean={self.mean}, "
            "scale={self.scale}, signed={self.signed})"
        )

    def estimate_distances(self, transformed_data, out=None, rescale=False):
        true_n, transformed_data = transformed_data
        if out is None:
            out = np.zeros(2 * len(transformed_data), dtype=np.uint64)
        estimate_pq_sse(transformed_data, self.tables, out, True)
        res = out.view(np.int8)
        res = res[:true_n]  # Trim padding elements
        if not rescale:
            # TODO: Would sorting actually be faster if we casted to float?
            return res
        # The center_norm is already built into res, so we just need to add
        # the norm of q and rescale.
        as_float = np.ascontiguousarray(res, dtype=np.float32)
        return self.q @ self.q + (as_float / self.scale + self.mean)

    def top(self, transformed_data, data, k=1, rescore=None):
        """
        Find the nearest data points to the query, given the compressed
        and non-compressed data to search.
        """
        true_n, transformed_data = transformed_data
        assert len(data) == true_n
        k = min(k, true_n)
        # In the first pass we collect `rescore` many rows
        if not rescore:
            rescore = min(2 * k + 10, true_n)
        assert true_n >= rescore >= k

        indices = np.zeros((rescore,), dtype=np.int64)
        values = np.zeros((rescore,), dtype=np.int32)
        init_heap(indices, values, True)
        query_pq_sse(transformed_data, true_n, self.tables, indices, values, True)

        # In a second pass we compute the true distances and return the actually
        # closest points. If we got fewer or exactly k outputs, there is no need
        # to compute the true distaneces.
        if rescore <= k:
            return indices, values

        # Remove padding from q
        diff = data[indices] - self.q[: data.shape[1]]
        dists = np.einsum("ij,ij->i", diff, diff)
        best = bottom_k(dists, k=k)
        return indices[best], dists[best]


class DummyPQ:
    def fit(self, data):
        return self

    def transform(self, data):
        return data

    def fit_transform(self, data):
        return self.fit(data).transform(data)

    def distance_table(self, q):
        return DummyDistanceTable(q)


class DummyDistanceTable:
    def __init__(self, q):
        self.q = q

    def estimate_distances(self, data, out=None, rescale=False):
        return ((data - self.q) * (data - self.q)).sum(axis=1)

    def top(self, transformed_data, data, k=1, rescore=None, out=None):
        dists = self.estimate_distances(transformed_data)
        best = bottom_k(dists, k)
        return best, dists[best]



import pytest
import numpy as np
import heapq

np.random.seed(10)

from fast_pq._fast_pq import estimate_pq_sse
from fast_pq._fast_pq import insert, init_heap

class Heap:
    def __init__(self, size):
        self.indices = np.empty((size,), dtype=np.int64)
        self.vals = np.empty((size,), dtype=np.int32)
        init_heap(self.indices, self.vals, signd=True)

    def insert(self, i, v):
        if v < self.peek():
            insert(self.indices, self.vals, i, v)

    def peek(self):
        return self.vals[0]


def test_heap_init():
    heap = Heap(3)
    np.testing.assert_array_equal(heap.indices, np.array([-1] * 3, dtype=np.int64))
    np.testing.assert_array_equal(heap.vals, np.array([127] * 3, dtype=np.int32))


def test_heap_insert_single_element():
    heap = Heap(1)
    heap.insert(1, 10)

    np.testing.assert_array_equal(heap.indices, np.array([1], dtype=np.int64))
    np.testing.assert_array_equal(heap.vals, np.array([10], dtype=np.int32))


def test_heap_insert_two_elements():
    heap = Heap(2)
    heap.insert(1, 10)
    np.testing.assert_array_equal(heap.indices, np.array([-1, 1], dtype=np.int64))
    np.testing.assert_array_equal(heap.vals, np.array([127, 10], dtype=np.int32))


def test_heap_unique():
    heap = Heap(2)
    heap.insert(1, 10)
    heap.insert(1, 10)
    np.testing.assert_array_equal(heap.indices, np.array([-1, 1], dtype=np.int64))
    np.testing.assert_array_equal(heap.vals, np.array([127, 10], dtype=np.int32))


def test_random():
    heap = Heap(10)
    pyheap = [(-127, -1)] * 10
    for t in range(1000):
        top_pyheap = -pyheap[0][0]
        assert top_pyheap == heap.peek()

        v = np.random.randint(10000 // (t+1))
        heap.insert(t, v)
        if v < top_pyheap:
            heapq.heappop(pyheap)
            heapq.heappush(pyheap, (-v, t))
        assert set(heap.vals) == set(-vi for vi, _ in pyheap)


def test_heap():
    np.random.seed(13)
    for n in range(1, 10):
        _test_sequence(list(range(n)))
        _test_sequence(list(reversed(range(n))))
        for _ in range(3):
            _test_sequence([np.random.randint(n) for _ in range(n)])


def _test_sequence(vs):
    heap = Heap(len(vs))
    verify_max_heap_property(heap.vals, 0)
    for i, v in enumerate(vs):
        heap.insert(i, v)
        assert v in heap.vals
        assert i in heap.indices
        verify_max_heap_property(heap.vals, 0)


def verify_max_heap_property(values, root):
    n = len(values)
    l, r = 2 * root + 1, 2 * root + 2
    if l < n:
        assert values[l] <= values[root]
        verify_max_heap_property(values, l)
    if r < n:
        assert values[r] <= values[root]
        verify_max_heap_property(values, r)



import numpy as np
from fast_pq import IVF, FastPQ, brute

np.random.seed(10)

n = 1000
d = 10
nq = 30
at = 10
dpb = 2

X = np.random.randn(n, d).astype(np.float32)
qs = np.random.randn(nq, d).astype(np.float32)

def compute_recall(metric, build_probes, query_probes):
    if at < n:
        trus = brute(qs, X, k=at, metric=metric)
    else:
        trus = np.broadcast_to(np.arange(n), (nq, n))
    ivf = IVF(metric, int(n**0.5), FastPQ(2))
    ivf.fit(X).build(X, n_probes=build_probes)
    recall_at = 0
    for q, tru in zip(qs, trus):
        guess = ivf.query(q, k=at, n_probes=query_probes)
        recall_at += len(set(guess) & set(tru))
    return recall_at / nq / at

def _test_monotone(metric):
    n = 4
    table = []
    for build_probes in range(1, n+1):
        table.append([])
        for query_probes in range(1, n+1):
            recall = compute_recall(metric, build_probes, query_probes)
            table[-1].append(recall)

    for row in table:
        print(row)

    # Test rows monotone
    for i in range(1, n):
        for j in range(n):
            assert table[i][j] >= table[i-1][j] - 0.1

    # Test cols monotone
    for i in range(n):
        for j in range(1, n):
            assert table[i][j] >= table[i][j-1] - 0.1

def test_euclidean():
    _test_monotone("euclidean")

def test_angular():
    _test_monotone("angular")




import numpy as np
import pytest

from fast_pq._fast_pq import estimate_pq_sse, query_pq_sse, init_heap
from fast_pq import FastPQ, cdist


def test_recall():
    np.random.seed(10)
    for i in range(1, 5):
        for method in ["argpartition", "top"]:
            n = np.random.randint(16 * i, 16 * (i + 1))
            recall_at_10 = _test_recall_inner(n, 8 * i, 100, 2, method)
            assert recall_at_10 > 0.8


def _test_recall_inner(n, d, k, dpb, method):
    X = np.random.randn(n, d).astype(np.float32)
    qs = np.random.randn(k, d).astype(np.float32)
    trus = cdist(qs, X).argmin(axis=1)

    pq = FastPQ(dims_per_block=dpb)
    data = pq.fit_transform(X)
    recall_at_10 = 0
    for q, tru in zip(qs, trus):
        dtable = pq.distance_table(q)
        if method == "argpartition":
            top10 = dtable.estimate_distances(data).argpartition(10)[:10]
        elif method == "top":
            top10, _ = dtable.top(data, X, 10)
        if tru in top10:
            recall_at_10 += 1
    return recall_at_10 / k


@pytest.mark.filterwarnings("ignore:Number of distinct clusters")
def test_topk():
    np.random.seed(11)
    for n in tuple(range(1, 10)) + (20, 30, 50):
        for dpb in [1, 2]:
            for signed in [True, False]:
                _test_topk_inner(n, 3, 11, dpb, signed)


@pytest.mark.filterwarnings("ignore:Number of distinct clusters")
def test_topk_0():
    np.random.seed(11)
    for signed in [True, False]:
        with pytest.raises(AssertionError):
            _test_topk_inner(0, 3, 11, 2, signed)


def test_fit_transform():
    np.random.seed(11)
    n, d = 100, 10
    X = np.random.randn(n, d).astype(np.float32)

    pq = FastPQ(2)
    n0, tdata0 = pq.fit_transform(X)
    n1, tdata1 = pq.transform(X)
    assert n0 == n1
    np.testing.assert_array_equal(tdata0, tdata1)


def _test_topk_inner(n, m, d, dpb, signed):
    X = np.random.randn(n, d).astype(np.float32)
    qs = np.random.randn(m, d).astype(np.float32)
    pq = FastPQ(dims_per_block=dpb)
    _, data = pq.fit_transform(X)

    for q in qs:
        dtable = pq.distance_table(q)

        out = np.zeros(2 * len(data), dtype=np.uint64)
        estimate_pq_sse(data, dtable.tables, out, signed)
        est = out.view(np.int8 if signed else np.uint8)
        est = est[:n] # Remove padding

        k = n
        indices = np.zeros((k,), dtype=np.int64)
        values = np.zeros((k,), dtype=np.int32)
        init_heap(indices, values, signed)
        query_pq_sse(data, n, dtable.tables, indices, values, signed)

        # Remove padding, and ignore
        # mask = indices < n |
        maxv = 127 if signed else 255
        mask = values < maxv  # we don't guarantee returing things with the max value
        indices, values = indices[mask], values[mask]
        values.sort()

        est.sort()
        est = est[est < maxv]
        assert np.all(est == values)


def test_large_labels():
    n, d, k = 100, 10, 100
    X = np.random.randn(n, d).astype(np.float32)
    q = np.random.randn(d).astype(np.float32)

    pq = FastPQ(2)
    _, data = pq.fit_transform(X)
    dtable = pq.distance_table(q)

    indices = np.empty((k,), dtype=np.int64)
    values = np.empty((k,), dtype=np.int32)
    labels = np.arange(n, dtype=np.int64) + 10**12
    init_heap(indices, values, True)
    query_pq_sse(data, n, dtable.tables, indices, values, True, labels)
    indices.sort()
    np.testing.assert_array_equal(indices, labels)




import numpy as np
import sklearn.cluster
from fast_pq import bottom_k, bottom_k_2d, FastPQ
from ._fast_pq import query_pq_sse, init_heap
import time
from contextlib import contextmanager


@contextmanager
def timer(verbose, text):
    if verbose:
        print(text)
        start = time.time()
    yield
    if verbose:
        print(f"Took {time.time() - start:.1f}s")


def cdist(X, Y, chunk=100):
    """
    Computes the squared Euclidean distances between two sets of points X and Y.
    Returns R st. R[i,j] = dist(X_i, Y_j)^2.
    Equivalent to scipy.spatial.distance.cdist.
    """
    # This is how sklearn computes row norms. It
    # %timeit np.linalg.norm(x, axis=1)
    # 486 ms ± 6.33 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
    # %timeit np.einsum('ij,ij->i',x,x)
    # 62.9 ms ± 2.35 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)
    # %timeit (x*x).sum(axis=1)
    # 495 ms ± 6.71 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
    nx = np.einsum("ij,ij->i", X, X)
    ny = np.einsum("ij,ij->i", Y, Y)
    res = np.zeros((nx.size, ny.size))
    for i in range(0, nx.size, chunk):
        res[i : i + chunk] = nx[i : i + chunk, None] + ny
        res[i : i + chunk] -= 2 * X[i : i + chunk] @ Y.T
    return res


def brute(X, Y, k, metric="euclidean", chunk=100):
    """
    Computes the k-nearest neighbors for each point in X based on the squared Euclidean distances
    between X and Y.
    """
    if metric == "angular":
        X = X / np.linalg.norm(X, axis=1, keepdims=True)
        Y = Y / np.linalg.norm(Y, axis=1, keepdims=True)
    elif metric not in ["angular", "euclidean"]:
        raise ValueError(f"Metric not supported: {metric}")
    n = X.shape[0]
    res = np.zeros((n, k))
    Ynorm2 = np.einsum("ij,ij->i", Y, Y)
    for i in range(0, n, chunk):
        Xchunk = X[i : i + chunk]
        Xnorm2 = np.einsum("ij,ij->i", Xchunk, Xchunk)
        part = Xnorm2[:, None] + Ynorm2[None] - 2 * Xchunk @ Y.T
        res[i : i + chunk] = part.argpartition(axis=1, kth=k)[:, :k]
    return res


def brute1(x, Y, k):
    """
    Computes the k-nearest neighbors for each point in X based on the squared Euclidean distances
    between X and Y.
    """
    diff = Y - x
    dists = np.einsum("ij,ij->i", diff, diff)
    return dists.argpartition(kth=k)[:k]


class IVF:
    def __init__(self, metric, n_clusters, pq=None):
        assert metric in ["euclidean", "angular"]
        self.metric = metric
        self.pq = FastPQ(dims_per_block=2) if pq is None else pq
        assert self.pq.centers is None, "PQ should not be pre-fitted"
        self.pq_transformed_points = [None] * n_clusters
        self.pq_transformed_centers = [None] * n_clusters
        self.n_clusters = n_clusters
        self.ids = [None] * n_clusters

    def fit(self, X, verbose=False):
        """
        Decides on IVF centers using full X, not PQ transformed.
        Doesn't insert any points into the data structure.
        If you want faster training, just fit on a sample rather than all the data
        like this: ivf.fit(X[np.random.choice(X.shape[0], 10**4, replace=False)]
        """

        n, d = X.shape
        assert n >= 1

        with timer(verbose, "Fitting IVF cluster centers..."):
            X = np.ascontiguousarray(X, dtype=np.float32)
            cl = sklearn.cluster.KMeans(
                n_clusters=self.n_clusters, n_init=1, verbose=verbose
            )

            if self.metric == "euclidean":
                cl.fit(X)
                self.all_centers = cl.cluster_centers_

            elif self.metric == "angular":
                # For angular we use kmeans clustering, but normalize the norm of the centers
                # as to make inner product equivalent to angular similarity.
                X = X / np.linalg.norm(X, axis=1, keepdims=True)
                cl.fit(X)
                self.all_centers = cl.cluster_centers_
                self.all_centers /= np.linalg.norm(self.all_centers, axis=1, keepdims=True)

        with timer(verbose, "Fitting PQ to data..."):
            self.pq.fit(X, verbose=verbose)

        return self

    def build(self, X, n_probes=2, verbose=False):
        """
        Builds the data structure by associating each data point in X with its nearest n_probes cluster centers.

        Parameters:
        -----------
        X : numpy.ndarray, shape (n_samples, n_features)
            The input data points to be indexed. Each row represents a data point with n_features dimensions.
        n_probes : int, optional, default: 2
            The number of nearest cluster centers to assign each data point to. Must be a positive integer.
        verbose : bool, optional, default: False
            If True, print additional information during the index building process.

        Returns:
        --------
        self : object
            Returns the instance itself, with the index structure built and data points assigned to the nearest n_probes cluster centers.
        """

        assert n_probes <= self.n_clusters, f"Can't assign points to {n_probes} clusters, as index only has {self.n_clusters}"
        self.data = data = X.copy()

        with timer(verbose, "Computing nearest clusters..."):
            # TODO: Use compression here, somehow.

            if self.metric == "euclidean":
                distances = cdist(data, self.all_centers)
            elif self.metric == "angular":
                data /= np.linalg.norm(data, axis=1, keepdims=True)
                distances = -data @ self.all_centers.T

            nearest_indices = bottom_k_2d(distances, n_probes)

        # We make sure that all centers have at least one point. This is just
        # a simple optimiation that's useful when there's not a lot of data
        # in the IVF, or if the query is OOD.
        with timer(verbose, "PQ Transforming active centers..."):
            self.active_centers = np.ascontiguousarray(
                self.all_centers[np.unique(nearest_indices)], dtype=np.float32
            )
            self.pq_transformed_centers = self.pq.transform(self.active_centers)

        t0, t1, t2 = 0, 0, 0
        with timer(verbose, "Transforming points and adding them to their nearest centers..."):
            for i in range(self.active_centers.shape[0]):
                # TODO: This is a lot slower than it needs to be.
                s = time.time()
                mask = np.any(nearest_indices == i, axis=1)
                t0 += time.time() - s

                # TODO: We should be able to extract the relevant data
                # directly from a previously transformed X.
                # However, that would require slicing into the compressed QuickADC format,
                # which is annoying.
                s = time.time()
                self.pq_transformed_points[i] = self.pq.transform(
                    np.ascontiguousarray(data[mask])
                )
                t1 += time.time() - s

                s = time.time()
                self.ids[i] = np.arange(data.shape[0], dtype=np.int64)[mask]
                t2 += time.time() - s

        if verbose:
            print(f"Finding any: {t0:.1f}")
            print(f"Transforming: {t1:.1f}")
            print(f"Computing ids: {t2:.1f}")

        return self

    def query(self, q, k, n_probes=1, pass_1=None):
        """
        Queries the data structure to find the top k closest elements to the given query vector q.

        Parameters
        ----------
        q : array-like
            The query vector for which to find the top k closest elements.
        k : int
            The number of closest elements to return.
        n_probes : int, optional, default=1
            The number of probes to use for searching the closest elements.

        Returns
        -------
        numpy.ndarray
            An array of indices corresponding to the top k closest elements to the query vector q.
        """

        q = np.ascontiguousarray(q, dtype=np.float32)
        if self.metric == "angular":
            q /= np.linalg.norm(q)
        dtable = self.pq.distance_table(q)

        # Find best centers
        top, _ = dtable.top(
            self.pq_transformed_centers, self.active_centers, k=n_probes
        )

        # For the first pass, get 2k candidates from each cluster
        # One may experiment with tuning this. Could even make it an argument to query.
        if pass_1 is None:
            pass_1 = (n_probes + 1) * k + 1
        indices = np.full(pass_1, -1, dtype=np.int64)
        values = np.full(pass_1, 127, dtype=np.int32)

        for i, cl in enumerate(top):
            true_n, transformed_data = self.pq_transformed_points[cl]
            query_pq_sse(
                transformed_data, true_n, dtable.tables, indices, values, True, labels=self.ids[cl]
            )

        # We may have gotten some heap padding elements mixed in. This is very rarely an issue,
        # so we check before we spend time building a mask and indexing.
        if -1 in indices:
            indices = indices[indices != -1]

        # If we have less or equal to k values, there's no point in rescoring
        if len(indices) <= k:
            return indices

        q = q[: self.data.shape[1]]  # Remove padding from q
        diff = self.data[indices] - q
        dists = np.einsum("ij,ij->i", diff, diff)
        best = bottom_k(dists, k=k)
        return indices[best]



import random
import math
import numpy as np
from functools import reduce

from fast_pq._fast_pq import estimate_pq_sse
from fast_pq._transform import transform_data, unpack, transform_tables


def test_simple():
    dat = [[1, 3, 7, 15]] + [[0, 0, 0, 0]] * 15
    tab = [list(range(16)) for i in range(4)]
    out = _slow_pq(
        np.array(dat).astype(np.uint8), np.array(tab).astype(np.uint8), signed=False
    )
    assert out[0] == sum((1, 3, 7, 15))
    assert all(out[i] == 0 for i in range(1, 16))


def test_rand():
    for i in range(1, 10):
        for j in range(1, 10):
            _test_rand_inner_signed(16 * i, 2 * j)
            _test_rand_inner_unsigned(16 * i, 2 * j)


def _test_rand_inner_unsigned(n, d):
    dat = [[random.randrange(16) for _ in range(d)] for _ in range(n)]
    # We saturate at 255, so no reason to have numbers much bigger than 1/d of that.
    tab = [[random.randrange(256 // d * 2) for _ in range(16)] for _ in range(d)]
    expected = np.array([sum(tab[j][dat[i][j]] for j in range(d)) for i in range(n)])
    expected = np.minimum(expected, 255)
    out = _slow_pq(np.array(dat).astype(np.uint8), np.array(tab).astype(np.uint8), False)
    assert np.all(expected == out)


def sat8_add(x, y):
    if x + y > 127:
        return 127
    if x + y < -128:
        return -128
    return x + y


def _test_rand_inner_signed(n, d):
    dat = [[random.randrange(16) for _ in range(d)] for _ in range(n)]
    # We saturate at 127, so no reason to have numbers much bigger than 1/sqrt(d) of that.
    top = int(math.floor(127 / d**0.5))
    tab = [[random.randrange(-top, top) for _ in range(16)] for _ in range(d)]
    # For unsigned addition saturation is associative, but for signed we have
    # to keep track of the stauration at every step.
    expected = np.array(
        [reduce(sat8_add, (tab[j][dat[i][j]] for j in range(d))) for i in range(n)]
    )
    out = _slow_pq(np.array(dat).astype(np.uint8), np.array(tab).astype(np.uint8), True)
    assert np.all(expected == out.astype(np.int8))


def _slow_pq(data0, tables0, signed):
    data = transform_data(data0)
    tables = transform_tables(tables0)
    out = np.zeros(2 * len(data), dtype=np.uint64)
    estimate_pq_sse(data, tables, out, signed)
    res = out.view(np.uint8)
    assert res.shape[0] == data0.shape[0]
    return res


def test_unpack():
    np.random.seed(10)
    n, d = 16 * 13, 2 * 7
    data = np.random.randint(16, size=(n, d)).astype(np.uint8)
    t_data = transform_data(data)
    data1 = unpack(t_data)
    np.testing.assert_array_equal(data, data1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, Extension
from Cython.Build import cythonize

setup(
    packages=["fast_pq"],
    ext_modules=cythonize(
        Extension(
            "fast_pq._fast_pq",
            sources=["fast_pq/_fast_pq.pyx"],
            extra_compile_args=[
                "-O3",
                "-march=native",
                "-ffast-math",
                #'-unroll-count=4',
                "-Wno-deprecated-declarations",
                "-Wno-deprecated-api",
                "-mavx",
                "-m64",
                #'-mprefetchwt1'
            ],
            language="c++",
        ),
        annotate=True,
    ),
)



import numpy as np
import pytest

from fast_pq import FastPQ, DummyPQ
from fast_pq import IVF, cdist, brute


def test_cdist():
    n1, n2, d = 10, 8, 5
    np.random.seed(10)
    for chunk in [1, 10, 100]:
        X = np.random.randn(n1, d)
        Y = np.random.randn(n2, d)
        dists = cdist(X, Y, chunk=chunk)
        for i in range(n1):
            for j in range(n2):
                tru_dist = np.sum((X[i] - Y[j]) ** 2)
                assert np.isclose(dists[i, j], tru_dist)


def test_brute():
    n1, n2, d = 40, 28, 5
    np.random.seed(10)
    X = np.random.randn(n1, d)
    Y = np.random.randn(n2, d)
    expected = cdist(X, Y).argpartition(axis=1, kth=10)[:, :10]
    best = brute(X, Y, 10)
    assert np.all(np.sort(expected) == np.sort(best))


def test_angular():
    n1, n2, d = 40, 28, 5
    np.random.seed(10)
    X = np.random.randn(n1, d)
    Y = np.random.randn(n2, d)
    X /= np.linalg.norm(X, axis=1, keepdims=True)
    Y /= np.linalg.norm(Y, axis=1, keepdims=True)
    # Ordering between euclidean and angular is the same for normalized vectors
    angular = brute(X, Y, 10, metric="angular")
    euclidean = brute(X, Y, 10, metric="euclidean")
    assert np.all(np.sort(angular) == np.sort(euclidean))


def test_small_n():
    d = 10
    for metric in ["euclidean", "angular"]:
        for n in range(1, 5):
            X = np.random.randn(n, d).astype(np.float32)
            q = np.random.randn(d).astype(np.float32)
            ivf = IVF(metric, 1, FastPQ(2))
            ivf.fit(X).build(X,n_probes=1)
            res = ivf.query(q, n)
            print(res)
            assert all(0 <= i < n for i in res)


def test_far_small_n():
    d = 10
    for metric in ["euclidean", "angular"]:
        for n in range(2, 5):
            X = np.random.randn(n, d).astype(np.float32)
            X[0, :] = 10**5
            q = np.random.randn(d).astype(np.float32)
            ivf = IVF(metric, 1, pq=FastPQ(2))
            ivf.fit(X).build(X, n_probes=1)
            res = ivf.query(q, n)
            print(res)
            assert all(0 <= i < n for i in res)


def test_euclidian_recall():
    np.random.seed(10)
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "euclidean", 1) > 0.1
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "euclidean", 2) > 0.2
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "euclidean", 4) > 0.35
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "euclidean", 8) > 0.50


def test_angular_recall():
    np.random.seed(10)
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "angular", 1) > 0.09
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "angular", 2) > 0.18
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "angular", 4) > 0.27
    assert _test_recall_inner(10**2, 20, 10, 2, 10, "angular", 8) > 0.36


def _test_recall_inner(n, d, nq, dpb, at, metric, n_probes):
    X = np.random.randn(n, d).astype(np.float32)
    qs = np.random.randn(nq, d).astype(np.float32)
    if at < n:
        trus = cdist(qs, X).argpartition(axis=1, kth=at)[:, :at]
    else:
        trus = np.broadcast_to(np.arange(n), (nq, n))
    ivf = IVF(metric, int(n**0.5), FastPQ(2))
    ivf.fit(X).build(X)
    recall_at = 0
    for q, tru in zip(qs, trus):
        guess = ivf.query(q, k=at, n_probes=n_probes)
        recall_at += len(set(guess) & set(tru))
    return recall_at / nq / at


@pytest.mark.filterwarnings("ignore:Number of distinct clusters")
def test_small():
    np.random.seed(10)
    assert _test_recall_inner(15, 10, 30, 2, 10, "euclidean", 1) > 0.05
